/*! \file Haar.cpp
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#if HAVE_CONFIG_H
#  include <config.h>
#endif // HAVE_CONFIG_H

#include <haar/Haar.h>
#include <util/Array2D.h>

#include <algorithm>

namespace haar
{

using namespace util;
using namespace math;
using std::string;
using std::vector;
using std::pair;
using namespace std;

//-----------------------------------------------------------//
//--------------------- 2D Haar stuff -----------------------//
//-----------------------------------------------------------//


void
Square2D::box (Box2f* box) const
{
    float invSize = 1.0f / (1 << l());

    float xf0 = x () * invSize;
    float yf0 = y () * invSize;
    float xf1 = (x () + 1) * invSize;
    float yf1 = (y () + 1) * invSize;
    
    box->min.set (xf0, yf0);
    box->max.set (xf1, yf1);
}


void
Haar2D::forwardTransform(Array2D<float>& image)
{
    int size = image.width();
    int lMax = iLog2 (size);
    
    // normalize input coefficients
    float wInv = 1.0f/size;
    for (int y = 0; y < size; y++)
        for (int x = 0; x < size; x++)
            image(x,y) = wInv*image(x,y);
    
    vector<float> tmp(size);
    for (int l = lMax; l > 0; l--)
    {
        int size  = 1 << l;
        int size2 = size >> 1;

        // split image horizontally
        for (int y = 0; y < size; y++)
        {
            for (int x = 0; x < size2; x++)
            {
                float a = image(2*x,y);
                float b = image(2*x+1, y);
                tmp[x]       = a + b;
                tmp[x+size2] = a - b;
            }
            for (int x = 0; x < size; x++)
                image(x,y) = tmp[x];
        }

        // split image vertically
        for (int x = 0; x < size; x++)
        {
            for (int y = 0; y < size2; y++)
            {
                float a = image(x,2*y);
                float b = image(x, 2*y+1);
                tmp[y]       = 0.5f*(a + b);
                tmp[y+size2] = 0.5f*(a - b);
            }
            for (int y = 0; y < size; y++)
                image(x,y) = tmp[y];
        }
    }
}


void
Haar2D::inverseTransform(Array2D<float>& image)
{
    int lMax = iLog2 (image.width ());
    vector<float> tmp(image.width());
    for (int l = 1; l <= lMax; l++)
    {
        int size  = 1 << l;
        int size2 = size >> 1;

        // split image horizontally
        for (int y = 0; y < size; y++)
        {
            for (int x = 0; x < size2; x++)
            {
                float a = image(x, y);
                float b = image(x+size2, y);
                tmp[2*x  ] = a + b;
                tmp[2*x+1] = a - b;
            }
            for (int x = 0; x < size; x++)
                image(x,y) = tmp[x];
        }

        // split image vertically
        for (int x = 0; x < size; x++)
        {
            for (int y = 0; y < size2; y++)
            {
                float a = image(x, y);
                float b = image(x, y+size2);
                tmp[2*y  ] = 0.5f*(a + b);
                tmp[2*y+1] = 0.5f*(a - b);
            }
            for (int y = 0; y < size; y++)
                image(x,y) = tmp[y];
        }
    }
    
    // normalize input coefficients
    for (int y = 0; y < image.width(); y++)
        for (int x = 0; x < image.width(); x++)
            image(x,y) = image.width()*image(x,y);
}


void
Haar2D::pack(const Array2D<float>& image, vector<pair<Key2D, float> >& coeffs)
{
    int size = image.width();
    int l = iLog2 (size) - 1;

    while (size > 1)
    {
        int size2 = size >> 1;
        // add in difference coefficients
        for (int y = 0; y < size2; y++)
        {
            for (int x = 0; x < size2; x++)
            {
                coeffs.push_back(pair<Key2D, float>(Key2D(l, x, y, 1), image(x + size2, y     )));
                coeffs.push_back(pair<Key2D, float>(Key2D(l, x, y, 2), image(x     , y + size2)));
                coeffs.push_back(pair<Key2D, float>(Key2D(l, x, y, 3), image(x + size2, y + size2)));
            }
        }

        l -= 1;
        size = size2;
    }
}


void
Haar2D::unPack(const Array2D<float>& image)
{
    // TODO
}

} // namespace haar
