/*! \file Haar.h
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef HAAR_HAAR_H
#define HAAR_HAAR_H

#include <util/Fwd.h>
#include <util/HashMap.h>
#include <math/Vec2.h>
#include <math/Box.h>
#include <vector>
#include <iostream>

namespace haar
{

inline int
iLog2 (float f)
{
    return (((*(int *)&f)&0x7f800000)>>23)-0x7f;
}

class Square2D;
class Key2D;

//-----------------------------------------------------------//
//--------------------- 2D Haar stuff -----------------------//
//-----------------------------------------------------------//

class Square2D
{
protected:
    char m_l;
    short m_x, m_y;

public:
    Square2D() : m_l(), m_x(), m_y() {}
    Square2D(int l, int x, int y) : m_l(l), m_x(x), m_y(y) {}
    Square2D(const Key2D& k);
    
    char  l() const {return m_l;}
    short x() const {return m_x;}
    short y() const {return m_y;}
    
    char&  l() {return m_l;}
    short& x() {return m_x;}
    short& y() {return m_y;}
    
    bool operator==(const Square2D& b) const
    {
        return l() == b.l() & y() == b.y() & x() == b.x();
    }
    
    void box(math::Box2f*) const;
    
    struct Hasher
    {
        size_t operator()(const Square2D& s) const
        {
            return s.y()*(1<<s.l()) + s.x();
        }
    };
};

class Key2D : public Square2D
{
protected:
    char m_t;

public:
    Key2D() : Square2D(), m_t() {}
    Key2D(int l, int x, int y, int t) : Square2D(l, x, y), m_t(t) {}
    Key2D(const Square2D& s, int t) : Square2D(s), m_t(t) {}

    char  t() const {return m_t;}
    
    bool operator==(const Key2D& b) const
    {
        return l() == b.l() & y() == b.y() & x() == b.x() & t() == b.t();
    }
    
    struct Hasher
    {
        size_t operator()(const Key2D& key) const
        {
            return key.t()*(1<<(2*key.l())) + key.y()*(1<<key.l()) + key.x();
        }
    };
};


inline Square2D::Square2D(const Key2D& s) :
    m_l(s.l()), m_x(s.x()), m_y(s.y()) 
{
    // empty
}


class Haar2D
{
public:
    typedef util::HashMap<Square2D, float, Square2D::Hasher> CSumMap;
    typedef util::HashMap<Key2D, float, Key2D::Hasher> CoeffMap;

    static void forwardTransform(util::Array2D<float>& image);
    static void inverseTransform(util::Array2D<float>& image);
    static void pack(const util::Array2D<float>& image,
                     std::vector<std::pair<Key2D, float> >&);
    static void unPack(const util::Array2D<float>& image);
};

inline std::ostream &
operator<<(std::ostream& out, const Square2D& v)
{
    return out << int (v.l()) << ", " << int (v.x()) << ", " << int (v.y());
}

inline std::ostream &
operator<<(std::ostream& out, const Key2D& v)
{
    return out << int (v.t()) << ", " << int (v.l()) << ", " << int (v.x()) << ", " << int (v.y());
}

} // namespace haar

#endif // HAAR_HAAR_H

