/*! \file WIS2D.cpp
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#if HAVE_CONFIG_H
#  include <config.h>
#endif // HAVE_CONFIG_H

#include <haar/WIS2D.h>
#include <haar/Wavelet2D.h>

namespace haar
{

using namespace util;
using namespace math;
using std::vector;


// local helper function definitions
namespace
{

inline void
warpAPoint (Vec2f* point, const Box2f& box, const Box2f& pBox)
{
    float xf = lerpFactor (pBox.min.x, pBox.max.x, point->x);
    float yf = lerpFactor (pBox.min.y, pBox.max.y, point->y);
    point->set (lerp (box.min.x, box.max.x, xf),
                lerp (box.min.y, box.max.y, yf));
}

inline void
setPBox (Box2f p[4], int i, float e[3][2])
{
    int ox = (i&1), oy = (i&2)>>1;
    p[i] = Box2f(Vec2f(e[0+ox][0], e[0+oy][1]), Vec2f(e[1+ox][0], e[1+oy][1]));
}

} // namespace 



void
WIS2D::setPoints (std::vector<Vec2f>& p)
{
    // set up pointers to points
    m_points.resize (p.size ());
    for (size_t i = 0; i < p.size (); ++i)
        m_points[i] = &p[i];
}

void
WIS2D::warpPoints () const
{
    warp (Square2D (0, 0, 0), m_wavelet->scale(), m_points,
          Box2f (Vec2f(0.0f), Vec2f(1.0f)));
}

void
WIS2D::warpPoint (Vec2f* p, float* pdf) const
{
    float scale = m_wavelet->scale ();
    Square2D s (0, 0, 0);
    Box2f pBox (Vec2f(0.0f), Vec2f(1.0f));
    Box2f pBoxes[4];
    int l, x, y;
    Box2f tempBox;

    while (true)
    {
        l = s.l ();
        x = s.x ();
        y = s.y ();
    
        if (l >= m_wavelet->numLevels())
        {
            s.box (&tempBox);
            warpAPoint (p, tempBox, pBox);
            break;
        }
    
        float prob[4];
        m_wavelet->childScales (scale, prob, s);
        
        for (int i = 0; i < 4; ++i) if (prob[i] < 0.0f) prob[i] = 0.0f;
    
        float midP[2], csum;
        float edge[3][2] =
            {
                {pBox.min[0], pBox.min[1]},
                {0, 0},
                {pBox.max[0], pBox.max[1]},
            };
        
        csum = 1.0f / (prob[0] + prob[1] + prob[2] + prob[3]);
        midP[1] = csum * (prob[0] + prob[1]);
        midP[0] = csum * prob[0] / midP[1];
        edge[1][0] = lerp (pBox.min[0], pBox.max[0], midP[0]);
        edge[1][1] = lerp (pBox.min[1], pBox.max[1], midP[1]);
        
        setPBox (pBoxes, 0, edge);
        setPBox (pBoxes, 1, edge);
        
        midP[0] = csum * prob[2] / (1.0f - midP[1]);
        edge[1][0] = lerp (pBox.min[0], pBox.max[0], midP[0]);
        
        setPBox (pBoxes, 2, edge);
        setPBox (pBoxes, 3, edge);
        
        int offset = 0;
        offset  = (p->y > pBoxes[offset].max.y) << 1;
        offset += (p->x > pBoxes[offset].max.x);
        
        s.l () = l + 1;
        s.x () = (x << 1) + int (bool (offset&1));
        s.y () = (y << 1) + int (bool (offset&2));
        pBox = pBoxes[offset];
        scale = prob[offset];
    }
}


void
WIS2D::warp (const Square2D& s, float scale,
             Points& points, const Box2f& pBox) const
{
    int l = s.l ();
    int x = s.x ();
    int y = s.y ();

    Box2f tempBox;

    if (l >= m_wavelet->numLevels())
    {
        ConstPointIterator itr = points.begin();
        for ( ; itr != points.end(); ++itr)
        {
            s.box (&tempBox);
            warpAPoint (*itr, tempBox, pBox);
        }
            
        return;
    }

    float prob[4];
    m_wavelet->childScales (scale, prob, s);
    
    for (int i = 0; i < 4; ++i) if (prob[i] < 0.0f) prob[i] = 0.0f;

    Box2f pBoxes[4];
    float midP[2], csum;
    float edge[3][2] =
        {
            {pBox.min[0], pBox.min[1]},
            {0, 0},
            {pBox.max[0], pBox.max[1]},
        };
    
    csum = 1.0f / (prob[0] + prob[1] + prob[2] + prob[3]);
    midP[1] = csum * (prob[0] + prob[1]);
    midP[0] = csum * prob[0] / midP[1];
    edge[1][0] = lerp (pBox.min[0], pBox.max[0], midP[0]);
    edge[1][1] = lerp (pBox.min[1], pBox.max[1], midP[1]);
    
    setPBox (pBoxes, 0, edge);
    setPBox (pBoxes, 1, edge);
    
    midP[0] = csum * prob[2] / (1.0f - midP[1]);
    edge[1][0] = lerp (pBox.min[0], pBox.max[0], midP[0]);
    
    setPBox (pBoxes, 2, edge);
    setPBox (pBoxes, 3, edge);
    
    
    // Reset point sets to size 0
    for (int i = 0; i < 4; ++i) m_pointCache[l][i].resize (0);
    
    ConstPointIterator itr = points.begin();
    for ( ; itr != points.end(); ++itr)
    {
        int offset = 0;
        offset += ((*itr)->y > pBoxes[offset].max.y) << 1;
        offset += ((*itr)->x > pBoxes[offset].max.x);
        m_pointCache[l][offset].push_back (*itr);
    }
    
    
    int l2 = l + 1;
    int x2 = x << 1;
    int y2 = y << 1;            
                  
    if (m_pointCache[l][0].size())
        warp (Square2D (l2, x2 + 0, y2 + 0), prob[0],
              m_pointCache[l][0], pBoxes[0]);
    
    if (m_pointCache[l][1].size())
        warp (Square2D (l2, x2 + 1, y2 + 0), prob[1],
              m_pointCache[l][1], pBoxes[1]);
    
    if (m_pointCache[l][2].size())
        warp (Square2D (l2, x2 + 0, y2 + 1), prob[2],
              m_pointCache[l][2], pBoxes[2]);
    
    if (m_pointCache[l][3].size())
        warp (Square2D (l2, x2 + 1, y2 + 1), prob[3],
              m_pointCache[l][3], pBoxes[3]);
}

} // namespace haar
