/*! \file Wavelet2D.cpp
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#if HAVE_CONFIG_H
#  include <config.h>
#endif // HAVE_CONFIG_H

#include <haar/Wavelet2D.h>
#include <util/Array2D.h>
#include <math.h>

#include <algorithm>

namespace haar
{

using namespace util;
using namespace math;
using std::string;
using std::vector;
using std::pair;
using namespace std;

// local helper function declaration
namespace
{
void addCoeffs2D(float result[ 4], float scale, const float coeffs[3]);
void pSum(float ca[4], float cb[4],
          const Wavelet2DBase& a, const Wavelet2DBase& b,
          const Square2D& p, Wavelet2DProduct::pSumMap& pSums);
bool smallerValue(const pair<Key2D, float>& a,
                  const pair<Key2D, float>& b)
{
    return fabs(a.second) > fabs(b.second);
}

} // namespace 


Wavelet2D::Wavelet2D(float p) :
    Wavelet2DBase(), m_numLevels(0), m_percent(p)
{

}


void
Wavelet2D::load(Array2D<float>& i)
{
    m_numLevels = iLog2 (i.width ());
    Haar2D::forwardTransform(i);
    
    m_scale = i(0,0);
    vector<pair<Key2D, float> > coeffs;
    coeffs.reserve(i.width()*i.height());
    Haar2D::pack(i, coeffs);
    sort(coeffs.begin(), coeffs.end(), smallerValue);
    
    int percentage = int(coeffs.size()*m_percent);
    
    m_coeffs.clear();
    m_coeffs.resize(percentage);
    for (int i = 0; i < percentage; i++)
        if (coeffs[i].second != 0.0f)
            m_coeffs[coeffs[i].first] = coeffs[i].second;
}


float
Wavelet2D::coeff(const Key2D& k) const
{
    Haar2D::CoeffMap::const_iterator i = m_coeffs.find(k);
    return (i == m_coeffs.end()) ? 0.0f : i->second;
}


void
Wavelet2D::childScales(float scale, float q[4],
                           const Square2D& s) const
{   
    int C_s = 1 << (s.l());
    float c[3] = {C_s*Wavelet2DBase::coeff(s, 1),
                  C_s*Wavelet2DBase::coeff(s, 2),
                  C_s*Wavelet2DBase::coeff(s, 3)};
    
    addCoeffs2D(q, scale, c);
}


bool
Wavelet2D::computeCSum(const Wavelet2DBase* b,
                           Haar2D::CSumMap& cSums) const
{
    cSums.clear();
    cSums.resize(m_coeffs.size());
    Haar2D::CoeffMap::const_iterator i;
    for (i = m_coeffs.begin(); i != m_coeffs.end(); ++i)
    {
        float p = b->coeff(i->first);
        if (p == 0.0f)
            continue;

        // multiply the two coefficients together
        p *= i->second;

        // add this sum to this node and all parent nodes
        for (Square2D s = i->first;
             s.l() >= 0;
             s.l()--, s.x()>>=1, s.y()>>=1)
        {
            Haar2D::CSumMap::iterator i = cSums.find(s);
            if (i != cSums.end())
                i->second += p;
            else
                cSums.insert(Haar2D::CSumMap::ValueType(s, p));
        }
    }
    return true;
}



Wavelet2DLerp::Wavelet2DLerp(const Wavelet2DBase* a,
                                     const Wavelet2DBase* b,
                                     const Wavelet2DBase* c,
                                     const Wavelet2DBase* d,
                                     float fx, float fy) :
    Wavelet2DBase(), m_a(a), m_b(b), m_c(c), m_d(d), m_fx(fx), m_fy(fy)
{
    m_scale = lerp(lerp(m_a->scale(), m_b->scale(), m_fx),
                   lerp(m_c->scale(), m_d->scale(), m_fx), m_fy);
    m_numLevels = min(min(m_a->numLevels(), m_b->numLevels()),
                      min(m_c->numLevels(), m_d->numLevels()));
}


float
Wavelet2DLerp::coeff(const Key2D& k) const
{
    float ca = m_a->coeff(k);
    float cb = m_b->coeff(k);
    float cc = m_c->coeff(k);
    float cd = m_d->coeff(k);

    return lerp(lerp(ca, cb, m_fx), lerp(cc, cd, m_fx), m_fy);
}


void
Wavelet2DLerp::childScales(float scale, float q[4],
                               const Square2D& s) const
{
    float ca[4], cb[4], cc[4], cd[4];
    m_a->childScales(scale, ca, s);
    m_b->childScales(scale, cb, s);
    m_c->childScales(scale, cc, s);
    m_d->childScales(scale, cd, s);
    
    for (int i = 0; i < 4; i++)
        q[i] = lerp(lerp(ca[i], cb[i], m_fx), lerp(cc[i], cd[i], m_fx), m_fy);
}


Wavelet2DProduct::Wavelet2DProduct(const Wavelet2DBase* a,
                                           const Wavelet2DBase* b) :
    Wavelet2DBase()
{
    setOperands(a, b);
}

void
Wavelet2DProduct::setOperands(const Wavelet2DBase* a,
                                  const Wavelet2DBase* b)
{
    m_a = a;
    m_b = b;
    
    cSum(*m_a, *m_b);
    Square2D p(0,0,0);
    
    float pa[3], pb[3];
    for (int i = 0; i < 3; i++)
    {
        pa[i] = m_a->coeff(p, i+1);
        pb[i] = m_b->coeff(p, i+1);
    }
    
    float qa[4], qb[4];
    addCoeffs2D(qa, m_a->scale(), pa);
    addCoeffs2D(qb, m_b->scale(), pb);
    
    m_pSums.clear();
    m_pSums.resize(500);
    m_pSums[p] = pair<children2D<float>, children2D<float> >(qa, qb);
                         
    m_scale = m_a->scale()*m_b->scale() + m_cSums[Square2D(0, 0, 0)];
    m_numLevels = std::min(m_a->numLevels(), m_b->numLevels());
}


inline void
Wavelet2DProduct::cSum(const Wavelet2DBase& m_a,
                           const Wavelet2DBase& m_b)
{
    m_cSums.clear();
    if (m_a.computeCSum(&m_b, m_cSums)) return;
    if (m_b.computeCSum(&m_a, m_cSums)) return;
    return;
}


float
Wavelet2DProduct::cSum(const Square2D& s) const
{
    Haar2D::CSumMap::const_iterator i = m_cSums.find(s);
    return (i == m_cSums.end()) ? 0.0f : i->second;
}



float
Wavelet2DProduct::coeff(const Key2D& k) const
{
    return 0.0f;
}


void
Wavelet2DProduct::childScales(float scale, float q[4],
                                  const Square2D& s) const
{
    float sa[4], sb[4];
    pSum(sa, sb, *m_a, *m_b, s, m_pSums);
    int C_s = 1 << (2*s.l()+2);
    int l = s.l() + 1;
    int x = s.x() << 1, y = s.y() << 1;
    
    q[0] = sa[0]*sb[0] + C_s*cSum(Square2D(l, x,   y  ));
    q[1] = sa[1]*sb[1] + C_s*cSum(Square2D(l, x+1, y  ));
    q[2] = sa[2]*sb[2] + C_s*cSum(Square2D(l, x,   y+1));
    q[3] = sa[3]*sb[3] + C_s*cSum(Square2D(l, x+1, y+1));
}


// local helper function definitions
namespace
{

void
addCoeffs2D(float q[4], float scale, const float c[3])
{
    q[0] = scale + c[0] + c[1] + c[2];
    q[1] = scale - c[0] + c[1] - c[2];
    q[2] = scale + c[0] - c[1] - c[2];
    q[3] = scale - c[0] - c[1] + c[2];
}

void
pSum(float ca[4], float cb[4],
     const Wavelet2DBase& a, const Wavelet2DBase& b,
     const Square2D& p, Wavelet2DProduct::pSumMap& pSums)
{
    // check for cached results
    Wavelet2DProduct::pSumMap::iterator i = pSums.find(p);
    if (i != pSums.end())
    {
        ca[0] = i->second.first.value[0];
        ca[1] = i->second.first.value[1];
        ca[2] = i->second.first.value[2];
        ca[3] = i->second.first.value[3];
        
        cb[0] = i->second.second.value[0];
        cb[1] = i->second.second.value[1];
        cb[2] = i->second.second.value[2];
        cb[3] = i->second.second.value[3];
        return;
    }

    // get parent's result
    float ppSumsA[4], ppSumsB[4];
    Square2D pp(p.l()-1, p.x() >> 1, p.y() >> 1);
    int qx = p.x() - 2*pp.x();
    int qy = p.y() - 2*pp.y();
    pSum(ppSumsA, ppSumsB, a, b, pp, pSums);
    float ppSumA = ppSumsA[(qy<<1) + qx];
    float ppSumB = ppSumsB[(qy<<1) + qx];
    
    // compute and cache the results
    int C_s = 1 << p.l();
    float pa[3], pb[3];
    for (int i = 0; i < 3; i++)
    {
        pa[i] = C_s*a.coeff(p, i+1);
        pb[i] = C_s*b.coeff(p, i+1);
    }
    
    addCoeffs2D(ca, ppSumA, pa);
    addCoeffs2D(cb, ppSumB, pb);

    pSums[p] = pair<children2D<float>, children2D<float> >(ca, cb);
}

} // namespace 


} // namespace haar
