/*! \file Wavelet2D.h
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef HAAR_WAVELET2D_H
#define HAAR_WAVELET2D_H

#include <haar/Haar.h>
#include <math/Vec2.h>

namespace haar
{

class Wavelet2DBase
{
public:
    Wavelet2DBase() : m_scale(1.0f) {}
    virtual ~Wavelet2DBase() {}

    virtual float coeff(const Key2D&) const = 0;
    virtual float coeff(const Square2D& s, int type) const
    {
        return coeff(Key2D(s, type));
    }
    
    virtual int numLevels() const = 0;
    virtual void childScales(float scale, float p[4],
                             const Square2D& s) const = 0;
    float scale() const {return m_scale;}
    
    virtual bool computeCSum(const Wavelet2DBase* wavelet,
                             Haar2D::CSumMap& cSum) const {return false;}
protected:
    float m_scale;
};


class Wavelet2D : public Wavelet2DBase
{
public:
    Wavelet2D(float p = 1.0f);
    virtual ~Wavelet2D() {}

    void setPercent (float p) {m_percent = p;}
    void load(util::Array2D<float>& array);
    virtual float coeff(const Key2D&) const;
    const Haar2D::CoeffMap& coeffs () const {return m_coeffs;}
    
    virtual int numLevels() const {return m_numLevels;}
    virtual void childScales(float scale, float p[4],
                           const Square2D& s) const;
    virtual bool computeCSum(const Wavelet2DBase* wavelet,
                             Haar2D::CSumMap& cSum) const;
protected:
    int m_numLevels;
    float m_percent;
    Haar2D::CoeffMap m_coeffs;
};


class Wavelet2DLerp : public Wavelet2DBase
{
public:
    Wavelet2DLerp() : Wavelet2DBase() {}
    Wavelet2DLerp(const Wavelet2DBase* a, const Wavelet2DBase* b,
                      const Wavelet2DBase* c, const Wavelet2DBase* d,
                      float fx, float fy);
            
    virtual int numLevels() const {return m_numLevels;}
    virtual void childScales(float scale, float p[4],
                             const Square2D& s) const;
    virtual float coeff(const Key2D&) const;

protected:
    const Wavelet2DBase* m_a, *m_b, *m_c, *m_d;
    float m_fx, m_fy;
    int m_numLevels;
};


template <typename T>
struct children2D
{
    children2D()
    {
        value[0] = value[1] = value[2] = value[3] = T();
    }
    children2D(const T& e1, const T& e2, const T& e3, const T& e4)
    {
        value[0] = e1;
        value[1] = e2;
        value[2] = e3;
        value[3] = e4;
    }
    children2D(const T v[4])
    {
        for (int i = 0; i < 4; i++)
            value[i] = v[i];
    }
        
    T value[4];
};

class Wavelet2DProduct : public Wavelet2DBase
{
public:
    Wavelet2DProduct() : Wavelet2DBase(), m_a(0), m_b(0) {}
    Wavelet2DProduct(const Wavelet2DBase* a,
                         const Wavelet2DBase* b);

    void setOperands(const Wavelet2DBase* a, const Wavelet2DBase* b);
    virtual float coeff(const Key2D&) const;
    virtual int numLevels() const {return m_numLevels;}
    virtual void childScales(float scale, float p[4],
                                     const Square2D& s) const;

    typedef util::HashMap<Square2D,
                          std::pair<children2D<float>,
                                    children2D<float> >,
                          Square2D::Hasher> pSumMap;
protected:
    void cSum(const Wavelet2DBase& a, const Wavelet2DBase& b);
    float cSum(const Square2D& s) const;
    
    Haar2D::CSumMap m_cSums;
    const Wavelet2DBase *m_a, *m_b;
    mutable pSumMap m_pSums;
    int m_numLevels;
};
                      
} // namespace haar
                      
#endif // HAAR_WAVELET2D_H
