/*! \file Box.h
    \brief Contains the definition of Box class.
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef MATH_BOX_H
#define MATH_BOX_H

#include <math/Fwd.h>
#include <math/Core.h>

namespace math
{

//! Represents a bounded interval in higher dimensions.
/*!
    Box is an N-D interval. It assumes that the template parameter Vec
    provides a dimensions() member (such as math::Vec2 or math::Vec3)
*/
template <typename Vec>
class Box
{
public:
    Vec min;            //!< The lower-bound of the interval
    Vec max;            //!< The upper-bound of the interval


    //-----------------------------------------------------------------------
    //@{ \name Constructors
    //-----------------------------------------------------------------------
    Box() {makeEmpty();}
    Box(const Vec & point) : min(point), max(point) {}
    Box(const Vec & minT, const Vec & maxT) : min(minT), max(maxT) {}
    template <typename T>
    Box(const Box<T> & box) : min((Vec)box.min), max((Vec)box.max) {}
    template <typename T, typename S>
    Box(const Box<T> & box1, const Box<S> & box2) :
        min((Vec)box1.min), max((Vec)box1.max)
    {
        enclose((Box<Vec>)box2);
    }
    //@}

    
    //-----------------------------------------------------------------------
    //@{ \name Equality and inequality
    //-----------------------------------------------------------------------
    bool operator== (const Box<Vec> & src) const
        {return (min == src.min && max == src.max);}
    bool operator!= (const Box<Vec> & src) const
        {return (min != src.min || max != src.max);}
    //@}

    
    //-----------------------------------------------------------------------
    //@{ \name Box manipulation
    //-----------------------------------------------------------------------
    void makeEmpty ()
    {
        min = Vec(10000000);
        max = Vec(-10000000);
    }
    void enclose (const Vec & point)
    {
        for (unsigned i = 0; i < Vec::dimensions(); i++)
        {
            min[i] = std::min(point[i], min[i]);
            max[i] = std::max(point[i], max[i]);
        }
    }
    void enclose (const Box<Vec> & box)
    {
        for (unsigned i = 0; i < Vec::dimensions(); i++)
        {
            min[i] = std::min(box.min[i], min[i]);
            max[i] = std::max(box.max[i], max[i]);
        }
    }
    void intersect (const Box<Vec> & box)
    {
        for (unsigned i = 0; i < Vec::dimensions(); i++)
        {
            min[i] = std::max(box.min[i], min[i]);
            max[i] = std::min(box.max[i], max[i]);
        }
    }
    void moveMinTo (const Vec &newMin)
    {
        Vec diff (newMin - min);
        min = newMin;
        max += diff;
    }
    void moveMaxTo (const Vec &newMax)
    {
        Vec diff (newMax - max);
        max = newMax;
        min += diff;
    }
    void makeValid ()
    {
        for (unsigned i = 0; i < Vec::dimensions(); i++)
        {
            if (min[i] > max[i])
                std::swap (min[i], max[i]);
        }
    }
    //@}

    
    //-----------------------------------------------------------------------
    //@{ \name Query functions
    //-----------------------------------------------------------------------
    Vec  size() const {return max-min;}
    Vec  center() const {return (max+min)/2;}
    bool contains (const Vec & point, bool proper = false) const
    {
        if (proper)
        {
            for (unsigned i = 0; i < Vec::dimensions(); i++)
                if (point[i] <= min[i] || point[i] >= max[i])
                    return false;
            return true;
        }
        else
        {
            for (unsigned i = 0; i < Vec::dimensions(); i++)
                if (point[i] < min[i] || point[i] > max[i])
                    return false;
            return true;
        }
    }
    bool intersects(const Box<Vec> & box) const
    {
        for (unsigned i = 0; i < Vec::dimensions(); i++)
            if (box.max[i] < min[i] || box.min[i] > max[i])
                return false;
        return true;
    }
    typename Vec::BaseType volume() const
    {
        typename Vec::BaseType retVal(1);
        Vec s = size();
        for (unsigned i = 0; i < Vec::dimensions(); i++)
            retVal *= s[i];
        return retVal;
    }
    typename Vec::BaseType area() const
    {
        typename Vec::BaseType retVal(0);
        Vec s = size();
        for (unsigned i = 0; i < Vec::dimensions(); i++)
            for (unsigned j = i+1; j < Vec::dimensions(); j++)
                retVal += s[i]*s[j];
        return 2*retVal;
    }
    unsigned majorAxis() const
    {
        unsigned major = 0;
        Vec s = size();
    
        for (unsigned i = 1; i < Vec::dimensions(); i++)
            if (s[i] > s[major])
                major = i;
        return major;
    }
    void boundingSphere (Vec* center, float* radius) const
    {
        *center = 0.5f * (min + max);
        *radius = intersects (*center) ? (*center - max).length () : 0.0f;
    }
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Classification
    //-----------------------------------------------------------------------
    bool hasVolume() const
    {
        for (unsigned i = 0; i < Vec::dimensions(); i++)
            if (max[i] <= min[i])
                return false;
        return true;
    }
    bool isEmpty() const
    {
        for (unsigned i = 0; i < Vec::dimensions(); i++)
            if (max[i] < min[i])
                return true;
        return false;
    }
    //@}
};


template <typename Vec>
inline std::ostream&
operator<<(std::ostream &o, const Box<Vec> &b)
{
    return o << "[(" << b.min << "), (" << b.max << ")]";
}


} // namespace math

#endif // MATH_BOX_H
