/*! \file Core.h
    \brief Contains the definition of various utility functions.
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef MATH_CORE_H
#define MATH_CORE_H

namespace math
{

//! Linear interpolation.
/*!
    Linearly interpolates between \a a and \a b, using parameter t.
    \param a A value.
    \param b Another value.
    \param t A blending factor of \a a and \a b.
    \return Linear interpolation of \a a and \b -
            a value between a and b if \a t is between 0 and 1.
*/
template <typename T, typename S>
inline T
lerp (T a, T b, S t)
{
    return (T) (a + t * (b - a));
}

//! Inverse linear interpolation.
template <typename T>
inline T
lerpFactor (T a, T b, T m)
{
    return (m - a) / (b - a);
}

//! Clamps a double between two bounds.
/*!
    \param a The value to test.
    \param l The lower bound.
    \param h The upper bound.
    \return The value \a a clamped to the lower and upper bounds.
*/
template <typename T>
inline T
clamp (T a, T l, T h)
{
    return (a < l) ? l : ((a <= h) ? a : h);
}

} // namespace math

// MATH_CORE_H
#endif
