/*! \file vec2.h
    \brief Contains the definition and implementation of the Vec2 class.
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef MATH_VEC2_H
#define MATH_VEC2_H

#include <math/Fwd.h>
#include <iostream>

namespace math
{

//! A general 2D vector class.
/*!
    This class handles storing and manipulating 2D vectors.
*/
template <typename T>
class Vec2
{
public:
    T x, y;
    
    //-----------------------------------------------------------------------
    //@{ \name Constructors and assignment
    //-----------------------------------------------------------------------
    //! Default constructor.
    /*!
        Call this constructor if you do not wish to initialize the vector to
        anything. It is more efficient than the others since it does nothing.
    */
    Vec2 () {}

    //! Parameter constructor.
    /*!
        Initialize the vector with the three parameters.
        \param a Value to set the x component to.
        \param b Value to set the y component to.
    */
    Vec2 (T a, T b) : x (a), y (b) {}

    //! Parameter constructor.
    /*!
        Initialize each component to \a a.
        \param a The value to initialize to.
    */
    Vec2 (T a) : x (a), y (a) {}
    
    //! Parameter constructor.
    /*!
        Initialize the vector from an array of doubles
        \param v The array of doubles to initialize from.
    */
    Vec2 (const T* v) : x (v[0]), y (v[1]) {}
    
    //! Copy constructor.
    /*!
        Initialize the vector to be a copy of \a v.
        \param v The vector to create a copy of.
    */
    template <typename S>
    Vec2 (const Vec2<S>& v) : x (T(v.x)), y (T(v.y)) {}

    //! Assignment operator.
    /*!
        Assigns the values from \a a to this Vec2.
    */
    const Vec2 & operator= (const Vec2& a) {x = a.x; y = a.y; return *this;}
    
    //! Assignment operator.
    /*!
        Sets all components of this Vec2 to \a a.
    */
    const Vec2 & operator= (T a) {x = y = a; return *this;}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Casting operators.
    //-----------------------------------------------------------------------
    //! Constant casting operator.
    /*!
        Casts this constant Vec2 to an array as a const T pointer.
    */
    operator const T* () const {return (const T*)&x;}
    
    //! Casting operator.
    /*!
        Casts this Vec2 to an array as a T pointer.
    */
    operator T* () {return (T*)&x;}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Element access and manipulation.
    //-----------------------------------------------------------------------
    //! Access operator.        
    /*!
        Returns the ith component of the vector.
        \param i The component to return.
        \warning i must be either 0, 1, or 2 in order to get expected results.
    */
    T & operator[] (int i) {return (&x)[i];}
    
    //! Constant access operator.
    /*!
        Returns the ith component of a constant vector.
        \param i The component to return.
        \warning i must be either 0, 1, or 2 in order to get expected results.
    */
    const T & operator[] (int i) const {return (&x)[i];}
    
    void set (T a) {x = y = a;}
    void set (T a, T b) {x = a; y = b;}
    void set (const Vec2 v) {x = v.x; y = v.y;}
    template <typename S>
    void set (const Vec2<S>& v) {x = T(v.x); y = T(v.y);}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Addition.
    //-----------------------------------------------------------------------
    //! Component-wise vector addition operator.
    Vec2 operator+ (const Vec2& v) const {return Vec2 (x + v.x, y + v.y);}
    
    //! Component-wise vector addition-assignment operator.
    const Vec2 & operator+= (const Vec2& v) {x += v.x; y += v.y; return *this;}

    //! Scalar addition-assignment operator.
    const Vec2 & operator+= (T a) {x += a; y += a; return *this;}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Subtraction.
    //-----------------------------------------------------------------------
    //! Component-wise vector subtraction operator.
    Vec2 operator- (const Vec2& v) const {return Vec2 (x - v.x, y - v.y);}
    
    //! Component-wise vector subtraction-assignment operator.
    const Vec2 & operator-= (const Vec2& v) {x -= v.x; y -= v.y; return *this;}
    
    //! Component-wise scalar subtraction assignment operator.
    const Vec2 & operator-= (T a) {x -= a; y -= a; return *this;}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Multiplication.
    //-----------------------------------------------------------------------
    //! Scalar multiplication operator.
    Vec2 operator* (T a) const {return Vec2 (x * a, y * a);}
    
    //! Component-wise vector multiplication operator.
    Vec2 operator* (const Vec2& v) const {return Vec2 (x * v.x, y * v.y);}
    
    //! Scalar multiplication-assignment operator.
    const Vec2 & operator*= (T a) {x *= a; y *= a; return *this;}
    
    //! Component-wise vector multiplication-assignment operator.
    const Vec2 & operator*= (const Vec2& v) {x *= v.x; y *= v.y; return *this;}
    
    //! Negation operator.
    Vec2 operator- () const {return Vec2 (-x, -y);}
    const Vec2 & negate () {x = -x; y = -y; return *this;}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Division.
    //-----------------------------------------------------------------------
    //! Scalar division operator.
    Vec2 operator/ (T a) const
    {
        T inv = T(1)/a;
        return Vec2 (x * inv, y * inv);
    }
    
    //! Component-wise vector division operator.
    Vec2 operator/ (const Vec2 & v) const {return Vec2 (x / v.x, y / v.y);}
    
    //! Scalar division-assignment operator.
    const Vec2 & operator/= (T a)
    {
        T inv = T(1)/a;
        x *= inv; y *= inv;
        return *this;
    }
    
    //! Component-wise vector division-assignment operator.
    const Vec2 & operator/= (const Vec2 & v) {x /= v.x; y /= v.y; return *this;}
    //@}

    static unsigned dimensions () {return 2;}
    typedef T    BaseType;
};


//! Multiply a scalar by a Vec2.
/*!
    \return The a Vec2 with each component multiplied by \a s.
    \relates Vec2
*/
template <typename S, typename T>
inline Vec2<T>
operator* (S s, const Vec2<T>& v)
{
    return Vec2<T> (v.x * s, v.y * s);
}


//! Output to a stream.
/*!
    Writes \a a to the output stream \a out, and returns the modified stream.
    \return A copy of \a out with the \a a written to it.
    \sa operator>>(std::istream&, Vec2&)
    \relates Vec2
*/
template <typename T>
inline std::ostream &
operator<< (std::ostream& out, const Vec2<T>& v)
{
    return out << v.x << " " << v.y ;
}


//! Input from a stream.
/*!
    Reads \a a from the input stream \a in, and returns the modified stream.
    \return A copy of \a in with the \a a already read from it.
    \sa operator<<(std::ostream&, const Vec2&)
    \relates Vec2
*/
template <typename T>
inline std::istream &
operator>> (std::istream& in, Vec2<T>& v)
{
    return in >> v.x >> v.y ;
}

} // namespace math

#endif // MATH_VEC2_H
