/*! \file Rand.h
    \brief Contains the implementation of various Pseudo-Random number
           generators.
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef RND_RAND_H
#define RND_RAND_H

#include <string.h>

//! Pseudo random number generators
namespace rnd
{

//! Common API for pseudo-random number generators.
/*!
    The base class provides the ability to create various random distributions
    using any derived class. The derived classes are only required to
    implement the next () function.
*/
class Rand
{
public:
    virtual ~Rand () {}
    virtual void init (unsigned long seed) = 0;

    //! Uniform bool in the range [false, true].
    bool          nextb ();
    
    //! Uniform byte in the range [0, 255].
    unsigned char nextc ();
    
    //! Uniform float in the range [0.0f, 1.0f).
    float         nextf ();
    
    //! Uniform float in the range [\a min, \a max).
    float         nextf (float min, float max);
    
    //! Uniform integer in the range [0, 2^32-1].
    unsigned long nexti ();
    
    //! Uniform integer in the range [\a min, \a max].
    long          nexti (long min, long max);
    
    //! Uniform double in the range [0.0, 1.0).
    double        nextd ();
    
    //! Uniform double in the range [\a min, \a max).
    double        nextd (double min, double max);
    
    //! STL RandomNumberGenerator Functor interface.
    long          operator () (long n);
    
protected:
    virtual unsigned long next () = 0;
};


inline bool
Rand::nextb ()
{
    return nexti () & 0x80000000UL;
}

inline unsigned char
Rand::nextc ()
{
    return (nexti () & 0xE0000000UL) >> 14;
}

inline float
Rand::nextf ()
{
    return nexti () * (1.0f/4294967296.0f);
}

inline float
Rand::nextf (float min, float max)
{
    return min + nextf ()* (max-min);
}

inline unsigned long
Rand::nexti ()
{
    return next ();
}

inline long
Rand::nexti (long min, long max)
{
    return min + long (nextf () * (max-min) + 0.5f);
}

inline double
Rand::nextd ()
{
    unsigned long a = nexti () >> 5, b = nexti () >> 6; 
    return (a*67108864.0+b)* (1.0/9007199254740992.0);
    //        2^26               2^53    
}

inline double
Rand::nextd (double min, double max)
{
    return min + nextd ()* (max-min);
}

inline long
Rand::operator () (long max)
{
    return long (nextf () * (max-1) + 0.5f);
}


//! Encapsulates a simple, fast PRNG
/*!
    This class encapsulates a fast PRNG as prescribed by Knuth, and described
    in Numerical Recipes.
*/
class RandKnuth : public Rand
{
public:
    RandKnuth (unsigned long seed = 0) {init (seed);}
    virtual ~RandKnuth () {}
    virtual void init (unsigned long seed);
    
protected:
    virtual unsigned long next ();

private:
    unsigned long m_state;
};


inline void
RandKnuth::init (unsigned long seed)
{
    m_state = (seed * 0xa5a573a5L) ^ 0x5a5a5a5aL;
}

inline unsigned long
RandKnuth::next ()
{
    m_state = 1664525L*m_state + 1013904223L;
    return m_state;
}

} // namespace rnd

#endif // RND_RAND_H
