/*! \file HashMap.h
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef UTIL_HASH_MAP_H
#define UTIL_HASH_MAP_H

#include <vector>
#include <functional>
#include <algorithm>
#include <iostream>

namespace util
{

template <typename Key>
class Hash
{
public:
    size_t operator()(const Key& key) const;
};

template <>
inline size_t
Hash<int>::operator()(const int& key) const
{
    return key;
}


// Note: assumes int is at least 32 bits.
enum { HASH_NUM_PRIMES = 28 };

static const unsigned hashPrimes[HASH_NUM_PRIMES] =
{
  53u,         97u,         193u,       389u,       769u,
  1543u,       3079u,       6151u,      12289u,     24593u,
  49157u,      98317u,      196613u,    393241u,    786433u,
  1572869u,    3145739u,    6291469u,   12582917u,  25165843u,
  50331653u,   100663319u,  201326611u, 402653189u, 805306457u, 
  1610612741u, 3221225473u, 4294967291u
};

inline unsigned hashNextPrime(unsigned n)
{
  const unsigned* first = hashPrimes;
  const unsigned* last = hashPrimes + (int)HASH_NUM_PRIMES;
  const unsigned* pos = std::lower_bound(first, last, n);
  return pos == last ? *(last - 1) : *pos;
}


template <typename Key,
          typename T,
          typename H = Hash<Key>,
          typename EQ = std::equal_to<Key> >
class HashMap
{
public:
    
    class iterator;
    class const_iterator;      

    typedef H                       Hasher;
    typedef Key                     KeyType;
    typedef T                       MappedType;
    typedef std::pair<const Key, T> ValueType;
    typedef EQ                      KeyEqual;
    
    typedef MappedType&             reference;
    typedef const MappedType&       const_reference;

    HashMap(const T& value = T(), size_t n = 10,
            const H& hash = H(), const EQ& e = EQ()) :
        m_defaultValue(value), m_hash(hash), m_numErased(0), m_eq(e)
    {
        size_t buckets = hashNextPrime(n);
        m_entries.reserve(buckets);
        m_buckets.resize(buckets, 0);
    }

    HashMap(const HashMap& hm) :
        m_entries(hm.m_entries),
        m_defaultValue(hm.m_defaultValue),
        m_hash(hm.m_hash),
        m_numErased(hm.m_numErased),
        m_eq(hm.m_eq)
    {
        deleteErased();
        m_entries.reserve(hm.m_entries.capacity());
        m_buckets.resize(hm.m_buckets.size(), 0);
        rehashAll();
    }
    
    HashMap& operator=(const HashMap& hm)
    {
        if (&hm != this)
        {
            m_entries.clear();
            m_buckets.clear();

            m_entries = hm.m_entries;
            deleteErased();
            m_entries.reserve(hm.m_entries.capacity());
            m_buckets.resize(hm.m_buckets.size(), 0);
            m_defaultValue = hm.m_defaultValue;
            m_hash = hm.m_hash;
            m_numErased = hm.m_numErased;
            m_eq = hm.m_eq;
            rehashAll();
        }
        return *this;
    }
    
    const_iterator begin() const
    {
        for (size_t i = 0; i < m_buckets.size(); i++)
            if (m_buckets[i])
                return const_iterator(this, m_buckets[i]);

        return end();
    }
    iterator begin()
    {
        for (size_t i = 0; i < m_buckets.size(); i++)
            if (m_buckets[i])
                return iterator(this, m_buckets[i]);

        return end();
    }
    
    const_iterator end() const {return const_iterator(this, 0);}
    iterator end() {return iterator(this, 0);}
    
    void insert(const ValueType& p)
    {
        if (m_entries.size() == m_entries.capacity())
            resize(m_entries.size()*2);
            
        size_t h = bucketNumber(p.first);

        m_entries.push_back(Entry(p.first, p.second, m_buckets[h]));
        m_buckets[h] = &m_entries.back();
    }
    
    void erase(iterator p)
    {
        if (!p.m_entry->erased)
            m_numErased++;
        p.m_entry->erased = true;
    }
    
    void erase(const Key& key)
    {
        iterator i = find(key);
        if (i != end())
            erase(i);
    }
    
    void clear()
    {
        m_entries.clear();
        m_buckets.clear();
        m_numErased = 0;
        resize(0);
        
    }
    
    reference operator[](const Key& key)
    {
        size_t h = bucketNumber(key);
        
        for (Entry* e = m_buckets[h]; e; e = e->next)
        {
            // found
            if (m_eq(e->data.first, key))
            {
                if (e->erased)  // re-insert
                {
                    e->erased = false;
                    m_numErased--;
                    return e->data.second = m_defaultValue;
                }
                return e->data.second;
            }
        }
        
        // not found

        if (m_entries.size() == m_entries.capacity())
        {
            resize(hashNextPrime(m_entries.size()*2));
            h = bucketNumber(key);
        }

        m_entries.push_back(Entry(key, m_defaultValue, m_buckets[h]));
        m_buckets[h] = &m_entries.back();
        return m_entries.back().data.second;
    }
    
    iterator find(const Key& key)
    {
        size_t h = bucketNumber(key);
        for (Entry* e = m_buckets[h]; e; e = e->next)
            if (m_eq(e->data.first, key) && !e->erased)
                return iterator(this, e);

        return iterator(this, 0);
    }
    
    const_iterator find(const Key& key) const
    {
        for (Entry* e = m_buckets[bucketNumber(key)]; e; e = e->next)
            if (m_eq(e->data.first, key) && !e->erased)
                return const_iterator(this, e);

        return const_iterator(this, 0);
    }
    
    size_t size() const {return m_entries.size()-m_numErased;}
    size_t bucket_count() const {return m_buckets.size();}
    void resize(size_t n)
    {
        deleteErased();
        
        size_t buckets = hashNextPrime(n);
        if (buckets <= m_buckets.size()) return;

        m_buckets.resize(buckets);
        std::fill(m_buckets.begin(), m_buckets.end(), (Entry*)0);
        m_entries.reserve(m_buckets.size());

        rehashAll();
    }
    
    size_t count(const Key& key) const
    {
        size_t c = 0;
        size_t h = bucketNumber(key);
        Entry* entry = m_buckets[h];
        while (entry)
        {
            entry = entry.next();
            if (!entry.erased())
                c++;
        }
        return c;
    }

private:
    
    void deleteErased()
    {
        size_t i = m_entries.size();
        while (m_numErased)
        {
            if (m_entries[--i].erased)
            {
                std::swap(m_entries[i], m_entries.back());
                m_entries.pop_back();
                --m_numErased;
            }
        }
    }
    
    void rehashAll()
    {
        for (size_t i = 0; i < m_entries.size(); i++)
        {
            size_t h = bucketNumber(m_entries[i].data.first);
            m_entries[i].next = m_buckets[h];
            m_buckets[h] = &m_entries[i];
        }
    }
    
    size_t bucketNumber(const Key& key) const
    {
        return m_hash(key) % m_buckets.size();
    }

    struct Entry
    {
        Entry(const Key& k = Key(), const T& v = T(), Entry* n = 0) :
            data(k, v), erased(false), next(n)
        {
            // empty
        }
        
        std::pair<Key, T> data;
        bool erased;
        Entry* next;
    };
    
    friend class iterator;
    friend class const_iterator;

    std::vector<Entry> m_entries;
    std::vector<Entry*> m_buckets;
    T m_defaultValue;
    Hasher m_hash;
    size_t m_numErased;
    EQ m_eq;
};


template <typename Key, typename T, typename H, typename EQ>
class HashMap<Key, T, H, EQ>::iterator
{
public:

    iterator() {}
    iterator(HashMap* hm, typename HashMap::Entry* e) : m_hm(hm), m_entry(e) {}
    
    iterator& operator++() 
    {
        const typename HashMap::Entry* old = m_entry;
        m_entry = m_entry->next;
        if (!m_entry)
        {
            size_t bucket = m_hm->bucketNumber(old->data.first);
            while (!m_entry && ++bucket < m_hm->m_buckets.size())
                m_entry = m_hm->m_buckets[bucket];
        }
        return *this;
    }
    
    iterator& operator++(int) 
    {
        iterator tmp = *this;
        ++*this;
        return tmp;
    }
    
    bool operator==(const iterator& it) const
    {
        return m_entry == it.m_entry;
    }
    
    bool operator!=(const iterator& it) const
    {
        return m_entry != it.m_entry;
    }
    
    std::pair<Key, T>& operator*()
    {
        return m_entry->data;
    }
    
    std::pair<Key, T>* operator->()
    {
        return &(operator*());
    }

private:
    friend class HashMap;
    HashMap* m_hm;
    typename HashMap::Entry* m_entry;
};


template <typename Key, typename T, typename H, typename EQ>
class HashMap<Key, T, H, EQ>::const_iterator
{
public:

    const_iterator() {}
    const_iterator(const iterator& it) : m_hm(it.m_hm), m_entry(it.m_entry) {}
    const_iterator(const HashMap* hm, const typename HashMap::Entry* e) : m_hm(hm), m_entry(e) {}
    
    const_iterator& operator++() 
    {
        const typename HashMap::Entry* old = m_entry;
        m_entry = m_entry->next;
        if (!m_entry)
        {
            size_t bucket = m_hm->bucketNumber(old->data.first);
            while (!m_entry && ++bucket < m_hm->m_buckets.size())
                m_entry = m_hm->m_buckets[bucket];
        }
        return *this;
    }
    
    const_iterator& operator++(int) 
    {
        const_iterator tmp = *this;
        ++*this;
        return tmp;
    }
    
    bool operator==(const const_iterator& it) const
    {
        return m_entry == it.m_entry;
    }
    
    bool operator!=(const const_iterator& it) const
    {
        return m_entry != it.m_entry;
    }
    
    const std::pair<Key, T>* operator->()
    {
        return &(operator*());
    }
    
    const std::pair<Key, T>& operator*() const
    {
        return m_entry->data;
    }

private:
    friend class HashMap;
    const HashMap* m_hm;
    const typename HashMap::Entry* m_entry;
};

} // namespace util

#endif // UTIL_HASH_MAP_H
