/*! \file Color.h
    \brief Contains the definition of Color classes
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#ifndef COLOR_H
#define COLOR_H

#include <iostream>

class Color3
{
protected:
    float r, g, b;

public:

    //-----------------------------------------------------------------------
    //@{ \name Constructors and assignment
    //-----------------------------------------------------------------------
    Color3 () {}
    Color3 (float x, float y, float z) : r (x), g (y), b (z) {}
    Color3 (float c) : r (c), g (c), b (c) {}
    Color3 (const float* c) : r (c[0]), g (c[1]), b (c[2]) {}
    const Color3 & operator= (float c) {r = g = b = c; return *this;}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Casting operators.
    //-----------------------------------------------------------------------
    operator const float* () const {return (const float*)&r;}
    operator float* () {return (float*)&r;}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Element access and manipulation.
    //-----------------------------------------------------------------------
    float& operator[](int i) {return (&r)[i];}
    const float & operator[](int i) const {return (&r)[i];}
    void set (float s) {r = g = b = s;}
    void set (float x, float y, float z) {r = x; g = y; b = z;}
    void set (const Color3& c) {r = c.r; g = c.g; b = c.b;}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Addition.
    //-----------------------------------------------------------------------
    Color3 operator+ (const Color3& c) const
    {
        return Color3 (r + c.r, g + c.g, b + c.b);
    }
    const Color3 & operator+= (const Color3& c)
    {
        r += c.r; g += c.g; b += c.b; return *this;
    }
    const Color3 & operator+= (float a)
    {
        r += a; g += a; b += a; return *this;
    }
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Subtraction.
    //-----------------------------------------------------------------------
    Color3 operator-(const Color3& c) const
    {
        return Color3 (r - c.r, g - c.g, b - c.b);
    }
    const Color3 & operator-=(const Color3& c)
    {
        r -= c.r; g -= c.g; b -= c.b; return *this;
    }
    const Color3 & operator-=(float a)
    {
        r -= a; g -= a; b -= a; return *this;
    }
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Multiplication.
    //-----------------------------------------------------------------------
    Color3 operator* (float a) const
    {
        return Color3 (r * a, g * a, b * a);
    }
    Color3 operator* (const Color3& c) const
    {
        return Color3 (r * c.r, g * c.g, b * c.b);
    }
    const Color3 & operator*= (float a)
    {
        r *= a; g *= a; b *= a; return *this;
    }
    const Color3 & operator*= (const Color3& c)
    {
        r *= c.r; g *= c.g; b *= c.b; return *this;
    }
    Color3 operator- () const {return Color3 (-r, -g, -b);}
    //@}


    //-----------------------------------------------------------------------
    //@{ \name Division.
    //-----------------------------------------------------------------------
    Color3 operator/ (float a) const
    {
        float inv = 1.0f / a;
        return Color3 (r * inv, g * inv, b * inv);
    }
    Color3 operator/ (const Color3 & c) const
    {
        return Color3 (r / c.r, g / c.g, b / c.b);
    }
    const Color3 & operator/= (float a)
    {
        float inv = 1.0f / a;
        r *= inv; g *= inv; b *= inv; return *this;
    }
    const Color3 & operator/= (const Color3 & c)
    {
        r /= c.r; g /= c.g; b /= c.b; return *this;
    }
    //@}
    
    float average () const
    {
        return (r + g + b) / 3.0f;
    }
    float luminance () const
    {
        return 0.212671f * r + 0.715160f * g + 0.072169f * b;
    }
    friend std::ostream& operator<< (std::ostream& out, const Color3& c)
    {
	   return (out << c.r << " " << c.g << " " << c.b);
    }
    friend std::istream& operator>> (std::istream& in, Color3& c)
    {
	   return (in >> c.r >> c.g >> c.b);
    }
    friend Color3 operator* (float s, const Color3& c)
    {
	   return Color3 (c.r * s, c.g * s, c.b * s);
    }
};


#endif
