/*! \file wisDemo.cpp
    \author Wojciech Jarosz
*/
/*
    Copyright (C) 2004-2006 by Wojciech Jarosz, University of California, San Diego.
    wjarosz@ucsd.edu

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the
    Free Software Foundation, Inc.,
    59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/
#if HAVE_CONFIG_H
#  include <config.h>
#endif // HAVE_CONFIG_H

#include GLUT_HEADER_LOCATION

#include <iostream>
#include <iomanip>
#include <math.h>

#include <util/Array2D.h>
#include <rnd/SequenceJittered.h>
#include <rnd/SequenceLP.h>
#include <haar/Wavelet2D.h>
#include <haar/WIS2D.h>
#include <EXRIO.h>

static char* usageString = "Wavelet Importance Sampling Demo\n"
                           "Copyright (c) 2004-2006, Wojciech Jarosz, "
                           "University of California, San Diego.\n"
                           "All rights reserved.\n"
                           "Usage: wisDemo image1 [image2] \n\n";

static const char* helpString = 
    "How to use this demo:\n"
    "\n"
    "  Keyboard usage is as follows:\n"
    "    '-'/'+'   Decrease/Increase number of points\n"
    "    '1'-'4'   Select random point sequence\n"
    "    '['/']'   Decrease/Increase % of Haar coeffs used (compression)\n"
    "    'l'       Reload image (use after '[' or ']')\n"
    "    't'       Toggle point/image view\n"
    "    'p'       Toggle viewing product\n"
    "    'w'       Toggle point warping\n";

using namespace std;
using namespace haar;
using namespace math;
using namespace rnd;
using namespace util;


inline bool isPowerOf2 (int v) {return (v & (v - 1)) == 0;}

enum
{
    RANDOM,
    UNIFORM,
    JITTERED,
    LARCHER_PILLICHSHAMMER,
    NUM_MODES
};

static const char *modes[] =
{
    "Random",
    "Uniform",
    "Jittered",
    "Larcher-Pillichshammer",
};

Jittered        random1 (1, 1, 1.0f);
Jittered        jittered (1, 1, 1.0f);
Jittered        uniform (1, 1, 0.0f);
LarcherPillichshammerBase2 lp (1);
Sequence*       currentGen = &lp;

Sequence* allGens[NUM_MODES] =
{
    &random1,
    &uniform,
    &jittered,
    &lp,
};

void display ();
void reshape (int w, int h);
void keyboard (unsigned char key, int x, int y);
void usage () {cout << usageString; exit (1);}
class WISDemo *window = 0;


class WISDemo
{
public:
    WISDemo (int argc, char*argv[],
             int setWidth, int setHeight,
             char * title);
    ~WISDemo ();

    void display ();
    void reshape (int w, int h);
    void keyboard (unsigned char key, int x, int y);
    
private:
    void makePoints ();
    void load ();
    void loadImage (Array2D<float>* array, const string& filename);
    void printStatus ();
    float scaleF (float x);
    float stepF (float x);
    float waveF (float xf, float yf, int type);
    void drawWavelet (const Wavelet2D& wavelet);
    void drawWaveletProd (const Wavelet2DBase& wavelet);


    int m_windowID;
    bool m_viewPoints, m_viewImportance;
    bool m_warp;
    bool m_viewProd;
    float m_percent;
    
    Wavelet2D m_wave1, m_wave2;
    Wavelet2DProduct m_prodWavelet;
    
    vector<Vec2f> m_wave1Points;
    vector<Vec2f> m_prodPoints;
    WIS2D m_wis1, m_wisProd;

    int m_mode, m_numPoints, m_sqrtNumPoints;
    
    string m_filename1, m_filename2;
};


WISDemo::WISDemo (int argc, char*argv[],
                  int width, int height,
                  char * title) :
    m_viewPoints (true),
    m_viewImportance (false),
    m_warp (true),
    m_viewProd (false),
    m_percent (1.0f),
    m_mode (LARCHER_PILLICHSHAMMER),
    m_numPoints (1024),
    m_sqrtNumPoints (int (sqrtf (m_numPoints)))
{
    glutInitDisplayMode (GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
    glutInitWindowSize (width, height);
    glutInitWindowPosition (0, 0);
    
    m_windowID = glutCreateWindow (title);
    glutDisplayFunc (::display);
    glutKeyboardFunc (::keyboard);
    glutReshapeFunc (::reshape);
    glutIdleFunc (0);

    // Background color
    glClearColor (0.0, 0.0, 0.0, 1.0);

    // Antialiasing settings
    glHint (GL_LINE_SMOOTH_HINT, GL_NICEST);
    glEnable (GL_BLEND);
    glBlendFunc (GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    glEnable (GL_LINE_SMOOTH);

    glPixelZoom (1.0f,-1.0f);
    
    if (argc > 2)
    {
        m_filename1 = argv[1];
        m_filename2 = argv[2];
    }
    else if (argc > 1)
    {
        m_filename1 = argv[1];
        m_filename2 = argv[1];
    }
    else
        usage ();

    load ();
    
    window = this;
}


WISDemo::~WISDemo ()
{
    glutDestroyWindow (m_windowID);
}


void
WISDemo::loadImage (Array2D<float>* array, const string& filename)
{
    Color3* image;
    int width, height;
    if (!(image = readEXRImage (filename, &width, &height)))
        usage ();

    int w = min (width, height);
    if (!isPowerOf2 (w) || width != height)
    {
        cerr << "image must be power of 2" << endl;
        usage ();
    }
    
    array->resizeErase (w, w);
    // copy into the array
    for (int y = 0; y < w; y++)
    {
        for (int x = 0; x < w; x++)
            (*array) (x, y) = image [x + w*y].luminance();
    }
    delete [] image;
}


void
WISDemo::load ()
{
    m_wave1.setPercent (m_percent);
    Array2D<float> i;
    loadImage (&i, m_filename1);
    m_wave1.load (i);
    
    m_wave2.setPercent (m_percent);
    loadImage (&i, m_filename2);
    m_wave2.load (i);
    
    m_prodWavelet = Wavelet2DProduct (&m_wave1, &m_wave2);

    m_wis1 = WIS2D (&m_wave1);
    m_wisProd = WIS2D (&m_prodWavelet);
    makePoints ();
}


void
WISDemo::makePoints ()
{
    m_prodPoints.clear ();
    m_wave1Points.clear ();
    
    random1.setNumSamples (1);
    jittered.setNumSamples (m_sqrtNumPoints);
    uniform.setNumSamples (m_sqrtNumPoints);
    lp.setNumSamples (m_numPoints);

    m_prodPoints.resize (m_numPoints);
    m_wave1Points.resize (m_numPoints);
    for (int i = 0; i < m_numPoints; i++)
    {
        currentGen->next (m_prodPoints[i]);
        m_wave1Points[i] = m_prodPoints[i];
    }
    
    m_wisProd.setPoints (m_prodPoints);
    m_wis1.setPoints (m_wave1Points);

    if (m_warp)
    {
        m_wisProd.warpPoints ();
        m_wis1.warpPoints ();
    }
}


inline float
WISDemo::scaleF (float x)
{
    return 1.0f;
}

inline float
WISDemo::stepF (float x)
{
    return x < 0.5f ? 1.0f : -1.0f;
}


inline float
WISDemo::waveF (float xf, float yf, int type)
{
    float retval = 1.0f;
    retval *= (type & 1) ? stepF (xf) : scaleF (xf);
    retval *= (type & 2) ? stepF (yf) : scaleF (yf);
    return retval;
}

void
WISDemo::drawWavelet (const Wavelet2D& wavelet)
{
    int l = wavelet.numLevels ();
    int w = 1 << l;
    Array2D<float> image (w, w);
    image.reset (wavelet.scale ());

    Haar2D::CoeffMap::const_iterator i;
    for (i = wavelet.coeffs ().begin (); i != wavelet.coeffs ().end (); ++i)
    {
        Key2D key (i->first);
        Box2f rangef;
        key.box (&rangef);
        Box2i range = Box2i (rangef.min*w, rangef.max*w);
        float p = (1 << key.l ());
        
        for (int y = range.min.y; y < range.max.y; y++)
        {
            float yf = (y - range.min.y)/float (range.size ().y);
            for (int x = range.min.x; x < range.max.x; x++)
            {
                float xf = (x - range.min.x)/float (range.size ().x);
                image (x,y) += waveF (xf,yf,key.t ()) * i->second*p;
            }
        }
    }

    for (int y = 0; y < w; y++)
        for (int x = 0; x < w; x++)
            image (x,y) = max (image (x,y), 0.0f);

    float v;
    Vec2f off (1.0f/w, 1.0f/w);
    for (int y = 0; y < w-1; y++)
        for (int x = 0; x < w-1; x++)
        {
            Box2f box1 (Vec2f (float(x)/w, float(y)/w),
                        Vec2f (float(x+1)/w, float(y+1)/w));

            v = image (x,y);
            glColor3f (v, v, v);
            glVertex2fv (2*box1.min - Vec2f (1.0f)+off);
            
            v = image (x+1,y);
            glColor3f (v, v, v);
            glVertex2fv (2*(box1.min+Vec2f (box1.size ().x, 0.0f)) - Vec2f (1.0f)+off);
            
            v = image (x+1,y+1);
            glColor3f (v, v, v);
            glVertex2fv (2*box1.max - Vec2f (1.0f)+off);
            
            v = image (x,y+1);
            glColor3f (v, v, v);
            glVertex2fv (2* (box1.min+Vec2f (0.0f, box1.size ().y)) - Vec2f (1.0f)+off);
        }
}

void
WISDemo::drawWaveletProd (const Wavelet2DBase& wavelet)
{
    int l = wavelet.numLevels ();
    int w = 1 << l;
    Array2D<float> image (w, w);
    for (int y = 0; y < w; y++)
    {
        for (int x = 0; x < w; x++)
        {
            float p[4];
            Square2D s (l-1, x >> 1, y >> 1);
            int qx = x - 2*s.x ();
            int qy = y - 2*s.y ();
            wavelet.childScales (1.0f, p, s);
            image (x,y) = max (p[ (qy<<1) + qx], 0.0f);
        }
    }

    float v;
    Vec2f off (1.0f/w, 1.0f/w);
    for (int y = 0; y < w-1; y++)
        for (int x = 0; x < w-1; x++)
        {
            Box2f box1 (Vec2f (float(x)/w, float(y)/w),
                        Vec2f (float(x+1)/w, float(y+1)/w));

            v = image (x,y);
            glColor3f (v, v, v);
            glVertex2fv (2*box1.min - Vec2f (1.0f)+off);
            
            v = image (x+1,y);
            glColor3f (v, v, v);
            glVertex2fv (2* (box1.min+Vec2f (box1.size ().x, 0.0f)) - Vec2f (1.0f)+off);
            
            v = image (x+1,y+1);
            glColor3f (v, v, v);
            glVertex2fv (2*box1.max - Vec2f (1.0f)+off);
            
            v = image (x,y+1);
            glColor3f (v, v, v);
            glVertex2fv (2* (box1.min+Vec2f (0.0f, box1.size ().y)) - Vec2f (1.0f)+off);
        }
}


void
WISDemo::display ()
{
    glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode (GL_PROJECTION);
    glLoadIdentity ();
    gluOrtho2D (-1.0f, 1.0f, -1.0f, 1.0f);
    
    glScalef (1.0f, -1.0f, 1.0f);
    if (m_viewImportance)
    {
        glBegin (GL_QUADS);
            if (m_viewProd)
                drawWaveletProd (m_prodWavelet);
            else
                drawWavelet (m_wave1);
        glEnd ();
    }
    
    if (m_viewPoints)
    {
        glColor3f (1.0f, 1.0f, 1.0f);
        glBegin (GL_POINTS);
        
            vector<Vec2f>* p = &m_wave1Points;
            if (m_viewProd)
                p = &m_prodPoints;

            for (size_t i = 0; i < p->size (); i++)
                glVertex2fv (2* (*p)[i] -  Vec2f (1.0f));
            
        glEnd ();
    }

    glutSwapBuffers ();
}


void
WISDemo::reshape (int width, int height)
{
    glViewport (0, 0, width, height);
    glMatrixMode (GL_MODELVIEW);
}


void
WISDemo::printStatus ()
{
    int width = 15;
    cout << "sample m_mode: " << setw (width) << modes[m_mode] <<
            " | num points: " << setw (5) << m_numPoints << ".\n";
}


void
WISDemo::keyboard (unsigned char key, int x, int y)
{
    if (key == 27)
    {
        glutDestroyWindow (m_windowID);
        exit (0);
    }
    
    int oldmode = m_mode;
    bool up = true;
    switch (key)
    {
        case '1': m_mode = RANDOM; break;
        case '2': m_mode = UNIFORM; break;
        case '3': m_mode = JITTERED; break;
        case '4': m_mode = LARCHER_PILLICHSHAMMER; break;    
    
        case '+': case '=':
            m_sqrtNumPoints +=1;
            m_sqrtNumPoints = (m_sqrtNumPoints > 0) ? m_sqrtNumPoints : 0;
            m_numPoints = m_sqrtNumPoints*m_sqrtNumPoints;
            makePoints ();
            break;
        case '-':
            m_sqrtNumPoints -=1;
            m_sqrtNumPoints = (m_sqrtNumPoints > 0) ? m_sqrtNumPoints : 0;
            m_numPoints = m_sqrtNumPoints*m_sqrtNumPoints;
            makePoints ();
            break;
        case 'l':
            load ();
            break;
        case '[':
            m_percent -= 0.01f;
            m_percent = max (m_percent, 0.001f);
            cout << "percentage = " << m_percent << endl;
            up = false;
            break;
        case ']':
            m_percent += 0.01f;
            m_percent = min (m_percent, 1.0f);
            cout << "percentage = " << m_percent << endl;
            up = false;
            break;

        case 't':
            if (m_viewImportance)
            {
                m_viewImportance = false;
                m_viewPoints = true;
            }
            else
            {
                m_viewImportance = true;
                m_viewPoints = false;
            }            
            break;
        case 'w':
            m_warp = !m_warp;
            makePoints ();
            break;
        case 'p':
            m_viewProd = !m_viewProd;
            break;
        break;
        default:
            up = false;
    }
    
    if (oldmode != m_mode)
    {
        currentGen = allGens[m_mode];
        makePoints ();
    }
    if (up)
    {
        printStatus ();
        display ();
    }
}


int
main (int argc, char*argv[])
{
    glutInit (&argc, argv);
    WISDemo demoWindow (argc, argv, 512, 512,
                        "Wavelet Importance Sampling Demo");
    cout << helpString;
    glutMainLoop ();
    
    return 0;
}


void display () {window->display();}
void reshape (int w, int h) {window->reshape(w, h);}
void keyboard (unsigned char key, int x, int y) {window->keyboard(key, x, y);}
